/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.client.cli;

import org.apache.flink.core.execution.CheckpointType;

import org.apache.commons.cli.CommandLine;

import static org.apache.flink.client.cli.CliFrontendParser.CHECKPOINT_FULL_OPTION;

/** Command line options for the CHECKPOINT command. */
public class CheckpointOptions extends CommandLineOptions {

    private final String[] args;

    private CheckpointType checkpointType;

    public CheckpointOptions(CommandLine line) throws CliArgsException {
        super(line);
        args = line.getArgs();
        // TODO: Support type INCREMENTAL
        if (line.hasOption(CHECKPOINT_FULL_OPTION.getOpt())) {
            checkpointType = CheckpointType.FULL;
        } else {
            checkpointType = CheckpointType.CONFIGURED;
        }
    }

    public String[] getArgs() {
        return args == null ? new String[0] : args;
    }

    public CheckpointType getCheckpointType() {
        return checkpointType;
    }
}
