/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.yarn.configuration;

import org.apache.flink.annotation.Internal;
import org.apache.flink.configuration.ConfigOption;

import static org.apache.flink.configuration.ConfigOptions.key;

/** Yarn configuration options that are not meant to be set by the user. */
@Internal
public class YarnConfigOptionsInternal {

    public static final ConfigOption<String> APPLICATION_LOG_CONFIG_FILE =
            key("$internal.yarn.log-config-file")
                    .stringType()
                    .noDefaultValue()
                    .withDescription(
                            "**DO NOT USE** The location of the log config file, e.g. the path to your log4j.properties for log4j.");
}
