/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.formats.hadoop.bulk.committer.cluster;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.hdfs.MiniDFSCluster;

import java.io.File;
import java.io.IOException;
import java.util.UUID;

/** Utility class for testing with HDFS FileSystem. */
public class HDFSCluster {

    public final MiniDFSCluster miniCluster;

    public HDFSCluster(File tmpDir) throws IOException {
        Configuration hdfsConfig = new Configuration();
        hdfsConfig.set(MiniDFSCluster.HDFS_MINIDFS_BASEDIR, tmpDir.getAbsolutePath());
        miniCluster = new MiniDFSCluster.Builder(hdfsConfig).build();
    }

    public void shutdown() {
        miniCluster.shutdown();
    }

    public Path newFolder() {
        return new Path(new Path(miniCluster.getURI() + "/"), UUID.randomUUID().toString());
    }
}
