/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.table.planner.plan.nodes.exec;

import org.apache.flink.FlinkVersion;

import java.util.List;

import static org.apache.flink.util.Preconditions.checkArgument;
import static org.apache.flink.util.Preconditions.checkNotNull;

/** The {@link ExecNodeGraph} representing the {@link ExecNode} topology. */
public class ExecNodeGraph {
    private final FlinkVersion flinkVersion;
    private final List<ExecNode<?>> rootNodes;

    public ExecNodeGraph(List<ExecNode<?>> rootNodes) {
        this(FlinkVersion.current(), rootNodes);
    }

    public ExecNodeGraph(FlinkVersion flinkVersion, List<ExecNode<?>> rootNodes) {
        this.flinkVersion = checkNotNull(flinkVersion, "The flinkVersion should not be null.");
        this.rootNodes = checkNotNull(rootNodes, "The rootNodes should not be null.");
        checkArgument(!rootNodes.isEmpty(), "The rootNodes should not be empty.");
    }

    public List<ExecNode<?>> getRootNodes() {
        return rootNodes;
    }

    public FlinkVersion getFlinkVersion() {
        return flinkVersion;
    }
}
