/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.runtime.checkpoint;

/** Status of the tracked checkpoint. */
public enum CheckpointStatsStatus {

    /** Checkpoint that is still in progress. */
    IN_PROGRESS,

    /** Checkpoint that has successfully completed. */
    COMPLETED,

    /** Checkpoint that failed. */
    FAILED;

    /**
     * Returns whether the checkpoint is in progress.
     *
     * @return <code>true</code> if checkpoint is in progress, <code>false</code> otherwise.
     */
    public boolean isInProgress() {
        return this == IN_PROGRESS;
    }

    /**
     * Returns whether the checkpoint has completed successfully.
     *
     * @return <code>true</code> if checkpoint has completed, <code>false</code> otherwise.
     */
    public boolean isCompleted() {
        return this == COMPLETED;
    }

    /**
     * Returns whether the checkpoint has failed.
     *
     * @return <code>true</code> if checkpoint has failed, <code>false</code> otherwise.
     */
    public boolean isFailed() {
        return this == FAILED;
    }
}
