[
{"namespace": "org.apache.flink.formats.parquet.generated",
 "type": "record",
 "name": "Address",
 "fields": [
     {"name": "num", "type": "int"},
     {"name": "street", "type": "string"},
     {"name": "city", "type": "string"},
     {"name": "state", "type": "string"},
     {"name": "zip", "type": "string"}
  ]
}
]
