/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.runtime.scheduler.adaptive.timeline;

import javax.annotation.Nullable;

import java.time.Instant;

/** Durable to record enter timestamp and leave timestamp. */
public class Durable {

    private final Long enterTimestamp;
    private @Nullable Long leaveTimestamp;

    public Durable() {
        this(Instant.now().toEpochMilli(), null);
    }

    public Durable(Long enterTimestamp, @Nullable Long leaveTimestamp) {
        this.enterTimestamp = enterTimestamp;
        this.leaveTimestamp = leaveTimestamp;
    }

    public void setLeaveTimestamp(long timestamp) {
        this.leaveTimestamp = timestamp;
    }

    public Long getEnterTimestamp() {
        return enterTimestamp;
    }

    public @Nullable Long getLeaveTimestamp() {
        return leaveTimestamp;
    }
}
