/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.runtime.io.network.api;

import org.apache.flink.annotation.Internal;

/**
 * Tells if the job is stopping because of consuming all data. {@link #DRAIN} means the job is
 * stopping either with a stop-with-savepoint --drain or due to consuming all records in the source.
 * A drained pipeline should call {@code StreamOperator#finish()} on all operators.
 */
@Internal
public enum StopMode {
    DRAIN,
    NO_DRAIN
}
