/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.runtime.scheduler.adaptive.allocator;

import org.apache.flink.runtime.clusterframework.types.ResourceID;
import org.apache.flink.runtime.jobmaster.slotpool.PhysicalSlot;

import javax.annotation.Nonnull;

/**
 * Helper class to represent the slot and the loading or slots utilization weight info of the task
 * executor where the slot is located at.
 */
class SlotTaskExecutorWeight<T> {
    final @Nonnull T taskExecutorWeight;
    final @Nonnull PhysicalSlot physicalSlot;

    SlotTaskExecutorWeight(@Nonnull T taskExecutorWeight, @Nonnull PhysicalSlot physicalSlot) {
        this.taskExecutorWeight = taskExecutorWeight;
        this.physicalSlot = physicalSlot;
    }

    ResourceID getResourceID() {
        return physicalSlot.getTaskManagerLocation().getResourceID();
    }
}
