/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.runtime.scheduler.taskexecload;

import org.apache.flink.annotation.Internal;

import java.util.Collection;
import java.util.List;
import java.util.stream.Collectors;

/**
 * The interface that holds the {@link TaskExecutionLoad} getter is required for corresponding
 * abstractions.
 */
@Internal
public interface HasTaskExecutionLoad {

    /**
     * Retrieves the task execution load.
     *
     * @return a {@link TaskExecutionLoad} instance
     */
    TaskExecutionLoad getTaskExecutionLoad();

    static <T extends HasTaskExecutionLoad> List<T> sortByTaskExecutionLoadDesc(Collection<T> c) {
        return c.stream()
                .sorted(
                        (leftReq, rightReq) ->
                                rightReq.getTaskExecutionLoad()
                                        .compareTo(leftReq.getTaskExecutionLoad()))
                .collect(Collectors.toList());
    }
}
