/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.streaming.api.functions.sink.legacy;

import org.apache.flink.annotation.Internal;
import org.apache.flink.api.common.SupportsConcurrentExecutionAttempts;

/**
 * A stream sink that ignores all elements.
 *
 * @param <T> The type of elements received by the sink.
 * @deprecated This interface will be removed in future versions. Use the new {@link
 *     org.apache.flink.streaming.api.functions.sink.v2.DiscardingSink} interface instead.
 */
@Internal
public class DiscardingSink<T> implements SinkFunction<T>, SupportsConcurrentExecutionAttempts {

    private static final long serialVersionUID = 1L;

    @Override
    public void invoke(T value) {}
}
