/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.streaming.api.operators;

import org.apache.flink.annotation.Experimental;
import org.apache.flink.api.common.operators.MailboxExecutor;
import org.apache.flink.streaming.api.operators.legacy.YieldingOperatorFactory;
import org.apache.flink.streaming.runtime.tasks.ProcessingTimeService;
import org.apache.flink.streaming.runtime.tasks.ProcessingTimeServiceAware;

import javax.annotation.Nullable;

import static org.apache.flink.util.Preconditions.checkNotNull;

/**
 * Base class for all stream operator factories. It implements some common methods and the {@link
 * ProcessingTimeServiceAware} interface which enables stream operators to access {@link
 * ProcessingTimeService}.
 */
@Experimental
public abstract class AbstractStreamOperatorFactory<OUT>
        implements StreamOperatorFactory<OUT>, ProcessingTimeServiceAware {

    protected ChainingStrategy chainingStrategy = ChainingStrategy.DEFAULT_CHAINING_STRATEGY;

    protected transient ProcessingTimeService processingTimeService;

    @Nullable private transient MailboxExecutor mailboxExecutor;

    @Override
    public void setChainingStrategy(ChainingStrategy strategy) {
        this.chainingStrategy = strategy;
    }

    @Override
    public ChainingStrategy getChainingStrategy() {
        return chainingStrategy;
    }

    @Override
    public void setProcessingTimeService(ProcessingTimeService processingTimeService) {
        this.processingTimeService = processingTimeService;
    }

    public void setMailboxExecutor(MailboxExecutor mailboxExecutor) {
        this.mailboxExecutor = mailboxExecutor;
    }

    /**
     * Provides the mailbox executor iff this factory implements {@link YieldingOperatorFactory}.
     */
    protected MailboxExecutor getMailboxExecutor() {
        return checkNotNull(
                mailboxExecutor, "Factory does not implement %s", YieldingOperatorFactory.class);
    }
}
