/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.table.types.inference.strategies;

import org.apache.flink.table.api.DataTypes;
import org.apache.flink.table.types.inference.TypeStrategiesTestBase;

import java.util.stream.Stream;

/** Tests for {@link ArrayAppendPrependTypeStrategy}. */
public class ArrayAppendPrependTypeStrategyTest extends TypeStrategiesTestBase {
    @Override
    protected Stream<TestSpec> testData() {
        return Stream.of(
                TestSpec.forStrategy(
                                "Array element is nullable, element to add is nullable",
                                SpecificTypeStrategies.ARRAY_APPEND_PREPEND)
                        .inputTypes(DataTypes.ARRAY(DataTypes.BIGINT()), DataTypes.BIGINT())
                        .expectDataType(DataTypes.ARRAY(DataTypes.BIGINT())),
                TestSpec.forStrategy(
                                "Array element type is non null, element to add is nullable",
                                SpecificTypeStrategies.ARRAY_APPEND_PREPEND)
                        .inputTypes(
                                DataTypes.ARRAY(DataTypes.BIGINT().notNull()), DataTypes.BIGINT())
                        .expectDataType(DataTypes.ARRAY(DataTypes.BIGINT())),
                TestSpec.forStrategy(
                                "Array element type is nullable, element to add is non null",
                                SpecificTypeStrategies.ARRAY_APPEND_PREPEND)
                        .inputTypes(
                                DataTypes.ARRAY(DataTypes.BIGINT()), DataTypes.BIGINT().notNull())
                        .expectDataType(DataTypes.ARRAY(DataTypes.BIGINT())),
                TestSpec.forStrategy(
                                "Array element type is non null, element to add is non null",
                                SpecificTypeStrategies.ARRAY_APPEND_PREPEND)
                        .inputTypes(
                                DataTypes.ARRAY(DataTypes.BIGINT().notNull()),
                                DataTypes.BIGINT().notNull())
                        .expectDataType(DataTypes.ARRAY(DataTypes.BIGINT().notNull())));
    }
}
