/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.table.planner.operations.converters.materializedtable;

import org.apache.flink.sql.parser.ddl.materializedtable.SqlAlterMaterializedTableDropDistribution;
import org.apache.flink.table.api.ValidationException;
import org.apache.flink.table.catalog.ObjectIdentifier;
import org.apache.flink.table.catalog.ResolvedCatalogMaterializedTable;
import org.apache.flink.table.catalog.TableChange;
import org.apache.flink.table.operations.Operation;
import org.apache.flink.table.operations.materializedtable.AlterMaterializedTableChangeOperation;

import java.util.List;
import java.util.function.Function;

/** A converter for {@link SqlAlterMaterializedTableDropDistribution}. */
public class SqlAlterMaterializedTableDropDistributionConverter
        extends AbstractAlterMaterializedTableConverter<SqlAlterMaterializedTableDropDistribution> {
    @Override
    protected Operation convertToOperation(
            SqlAlterMaterializedTableDropDistribution sqlDropDistribution,
            ResolvedCatalogMaterializedTable oldTable,
            ConvertContext context) {
        ObjectIdentifier identifier = resolveIdentifier(sqlDropDistribution, context);
        if (oldTable.getDistribution().isEmpty()) {
            throw new ValidationException(
                    String.format(
                            "Materialized table %s does not have a distribution to drop.",
                            identifier));
        }

        return new AlterMaterializedTableChangeOperation(
                identifier, gatherTableChanges(sqlDropDistribution, context), oldTable);
    }

    @Override
    protected Function<ResolvedCatalogMaterializedTable, List<TableChange>> gatherTableChanges(
            SqlAlterMaterializedTableDropDistribution sqlAlterTable, ConvertContext context) {
        return oldTable -> List.of(TableChange.dropDistribution());
    }
}
