/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.flink.table.planner.plan.stream.sql.join

import org.apache.flink.table.planner.plan.common.JoinReorderTestBase
import org.apache.flink.table.planner.utils.TableTestUtil
import org.apache.flink.testutils.junit.extensions.parameterized.ParameterizedTestExtension

import org.junit.jupiter.api.extension.ExtendWith

/** The plan test for join reorder in stream mode. */
@ExtendWith(Array(classOf[ParameterizedTestExtension]))
class JoinReorderTest(isBushyJoinReorder: Boolean)
  extends JoinReorderTestBase(isBushyJoinReorder: Boolean) {
  override protected def getTableTestUtil: TableTestUtil = streamTestUtil()
}
