/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.formats.protobuf;

import org.apache.flink.formats.protobuf.testproto.RepeatedTest;
import org.apache.flink.table.data.ArrayData;
import org.apache.flink.table.data.RowData;

import org.junit.jupiter.api.Test;

import static org.assertj.core.api.Assertions.assertThat;

/** Test conversion of proto list of primitive data to flink internal data. */
class RepeatedProtoToRowTest {
    @Test
    void testRepeated() throws Exception {
        RepeatedTest simple = RepeatedTest.newBuilder().setA(1).addB(1).addB(2).build();
        RowData row = ProtobufTestHelper.pbBytesToRow(RepeatedTest.class, simple.toByteArray());

        assertThat(row.getArity()).isEqualTo(6);
        assertThat(row.getInt(0)).isEqualTo(1);
        ArrayData arr = row.getArray(1);
        assertThat(arr.size()).isEqualTo(2);
        assertThat(arr.getLong(0)).isEqualTo(1L);
        assertThat(arr.getLong(1)).isEqualTo(2L);
    }
}
