/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.runtime.entrypoint;

import javax.annotation.Nonnull;

import java.util.Properties;

/**
 * Configuration class which contains the parsed command line arguments for the {@link
 * ClusterEntrypoint}.
 */
public class ClusterConfiguration {

    @Nonnull private final String configDir;

    @Nonnull private final Properties dynamicProperties;

    @Nonnull private final String[] args;

    public ClusterConfiguration(
            @Nonnull String configDir,
            @Nonnull Properties dynamicProperties,
            @Nonnull String[] args) {
        this.configDir = configDir;
        this.dynamicProperties = dynamicProperties;
        this.args = args;
    }

    @Nonnull
    public String getConfigDir() {
        return configDir;
    }

    @Nonnull
    public Properties getDynamicProperties() {
        return dynamicProperties;
    }

    @Nonnull
    public String[] getArgs() {
        return args;
    }
}
