/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.table.planner.plan.logical;

import org.apache.flink.table.runtime.operators.window.tvf.common.WindowAggOperator;

import org.apache.flink.shaded.jackson2.com.fasterxml.jackson.annotation.JsonIgnore;
import org.apache.flink.shaded.jackson2.com.fasterxml.jackson.annotation.JsonSubTypes;
import org.apache.flink.shaded.jackson2.com.fasterxml.jackson.annotation.JsonTypeInfo;

/** Logical representation of a window specification. */
@JsonTypeInfo(use = JsonTypeInfo.Id.NAME, include = JsonTypeInfo.As.PROPERTY, property = "type")
@JsonSubTypes({
    @JsonSubTypes.Type(value = TumblingWindowSpec.class),
    @JsonSubTypes.Type(value = HoppingWindowSpec.class),
    @JsonSubTypes.Type(value = CumulativeWindowSpec.class),
    @JsonSubTypes.Type(value = SessionWindowSpec.class)
})
public interface WindowSpec {

    String toSummaryString(String windowing, String[] inputFieldNames);

    /**
     * Return true if the window is aligned.
     *
     * <p>See more details about aligned window and unaligned window in {@link WindowAggOperator}.
     */
    @JsonIgnore
    boolean isAlignedWindow();
}
