/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.streaming.api.operators.python.embedded;

/** Test-only iterator that can be loaded through an isolated classloader. */
public class ForeignClassLoaderIterator {

    private final Object[] values;
    private int index;
    private boolean closed;

    public ForeignClassLoaderIterator(Object... values) {
        this.values = values;
    }

    public boolean hasNext() {
        return index < values.length;
    }

    public Object next() {
        return values[index++];
    }

    public void close() {
        closed = true;
    }

    public boolean isClosed() {
        return closed;
    }
}
