/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.table.catalog;

import org.apache.flink.table.api.TableException;

import org.junit.jupiter.api.Test;

import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.assertThatThrownBy;

/** Tests for {@link ObjectIdentifier}. */
class ObjectIdentifierTest {

    @Test
    void testAnonymousIdentifier() {
        String objectName = "my_anonymous_table";
        ObjectIdentifier objectIdentifier = ObjectIdentifier.ofAnonymous(objectName);

        assertThat(objectIdentifier.getCatalogName()).isEqualTo(ObjectIdentifier.UNKNOWN);
        assertThat(objectIdentifier.getDatabaseName()).isEqualTo(ObjectIdentifier.UNKNOWN);
        assertThat(objectIdentifier.toList()).containsExactly(objectName);
        assertThat(objectIdentifier.asSummaryString())
                .isEqualTo(objectIdentifier.toString())
                .isEqualTo(objectName);

        assertThatThrownBy(objectIdentifier::asSerializableString)
                .isInstanceOf(TableException.class);
        assertThatThrownBy(objectIdentifier::toObjectPath).isInstanceOf(TableException.class);
    }
}
