/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.table.functions;

import org.apache.flink.annotation.Internal;
import org.apache.flink.table.connector.ChangelogMode;
import org.apache.flink.table.functions.ChangelogFunction.ChangelogContext;
import org.apache.flink.table.types.inference.TypeStrategy;

/**
 * Strategy for determining the output {@link ChangelogMode} of a built-in process table function.
 *
 * <p>Similar to {@link TypeStrategy}, this is used to declare changelog semantics in the function
 * definition rather than implementing the {@link ChangelogFunction} interface.
 */
@Internal
public interface ChangelogModeStrategy {

    /** Infers the output {@link ChangelogMode} based on the given {@link ChangelogContext}. */
    ChangelogMode inferChangelogMode(ChangelogContext changelogContext);
}
